from luca.model import notfound, sqlerror, log, function_points_nowild, PermissionError
from luca import model

########################################################## ACCOUNT MASK CHECKING

def account_level(account):
	'''Account level (top level = 1'''
	return len(account.split(mask_separator))

def check_chart(chart_row, checkdigit):
	'''Checks whether a chart of accounts is hierarchically correct.
	Returns a sorted list of accounts (with individual account problems pointed out)
	plus the first error found in chart (or empty if it is completely OK)

	It also fills up the 'css_class' which determines the appearance of chart.

	This function handles *NO* SQL errors, so the chart_row must be valid!
	It is a good idea to enclose the call in a try..except sqlerror 
	'''

	first_error = ""

	mask = chart_row.mask

	# paranoid
	err = check_chart_mask(mask)
	if err:
		# fatal error, give up early
		return ([], err)

	max_level = account_level(mask)

	# Copies all accounts to a local dictionary to re-check it off-line
	daccounts = {}
	laccounts = []
	i = 0
	for account in chart_row.accounts:
		# paranoid mask checks
		err = check_account_mask_naked(mask, account.code, account.aggregate)
		first_error = first_error or err

		# Calculates account aggregation level (lower level -> higher aggregation)
		level = account_level(account.code)
		
		if not err:
			if account.parent:
				if level == 1:
					err = "Toplevel should not have parent"
				else:
					err = check_account_mask_naked(mask, account.parent, 1)
			elif level > 1:
				err = "Has no parent"

		first_error = first_error or err
		
		css_class = i % 2 and "odd" or "even"
		if level <= 1:
			css_class += " accounttoplevel"
		elif account.aggregate:
			css_class += " accountaggregate"

		parent_class = collapse_class = account_to_class(account.parent)	
		code_class = account_to_class(account.code)
		if level <= 1:
			is_leaf = 0
			collapse_class += " collapsed"
		elif level >= max_level:
			is_leaf = 1 
			collapse_class += " invisible"
		else:
			is_leaf = 0
			collapse_class += " collapsed invisible"

	
		d = {"code": account.code, "parent": account.parent, "aggregate": account.aggregate,
		     "xclass": account.xclass, "level": level, "id": account.id, "err": err, "css_class": css_class,
			"collapse_class":  collapse_class, "parent_class": parent_class, "code_class": code_class,
			"description": account.description, "mobile": account.mobile, "inactive": account.inactive, 
			"leaf": is_leaf}

		daccounts[account.code] = d
		laccounts.append(d)
		i += 1
	
	# guarantees until now (if first_error is empty):
	# * Account masks are ok
	# * Top-level accounts are guaranteed to have level=1
	# * Non-top-levels are guaranteed to have level > 1

	for account in laccounts:

		if not account["parent"]:
			# top-level aggregate account
			continue

		# Levels 2 and beyond: check if parent exists
		if account["parent"] not in daccounts:
			account["err"] = "Parent %s does not exist" % account["parent"]
			first_error = first_error or account["err"]
			continue

		parent = daccounts[account["parent"]]

		# Checks whether the parent is more aggregate (lower level) than daughter
		if account["level"] <= parent["level"]:
			account["err"] = "More aggregate than its parent (%s)" % account["parent"]
			first_error = first_error or account["err"]
			continue

		# Parents must always be aggregate
		if not parent["aggregate"]:
			account["err"] = "Parent (%s) is not aggregate" % account["parent"]
			first_error = first_error or account["err"]
			continue

		# Class of parent and non-aggregate daughter must be the same
		if not account["aggregate"] and (account["xclass"] != parent["xclass"]):
			account["err"] = "Account and parent have different classes"
			first_error = first_error or account["err"]
			continue

		if account["aggregate"] and account["level"] >= max_level:
			account["err"] = "Accordingly to code, should not be aggregate"
			first_error = first_error or account["err"]
			continue

		if (not account["aggregate"]) and account["level"] < max_level:
			account["err"] = "Accordingly to code, should be aggregate" 
			first_error = first_error or account["err"]
			continue

	# Guarantees (if first_error is empty):
	#
	# * Accounts that have daughters are always aggregate
	# * non-aggregate accounts must fill the whole code-mask (and vice-versa)
	# * parent accounts are always more aggregate than the daughters
	# * - guarantees that there are no loops in the hierarchy
	# * All non-top-level accounts have a parent
	# * - guarantees that the chart tree is correct

	if first_error:
		for account in laccounts:
			if account["err"]:
				account["css_class"] = "accounterr"

	for account in laccounts:
		if checkdigit and not account["aggregate"]:
			account["code_checkdigit"] = add_checkdigit(account["code"])
		else:
			account["code_checkdigit"] = account["code"]

	laccounts.sort(lambda x,y: cmp(x["code"], y["code"]))

	return (laccounts, first_error)

def account_to_class(code):
	return "par" + code.replace(".", "_")

def add_checkdigit(code):
	'''Returns any alphanumeric code with checkdigit suffix'''
	return code + checkdigit_separator + checkdigit(code)

def checkdigit(code):
	'''Returns the check digit for a given code (account code or any other alphanumeric code). 
	   Loosely inspired in modulo 11, with provision for non-numeric characters'''
	sum = 0
	multiplier = 12
	for c in code:
		sum += ord(c) * multiplier
		multiplier += 1
		if multiplier >= 19:
			multiplier = 13
	checkdigit = 11 - (sum % 11)
	if checkdigit >= 10:
		checkdigit = 0
	return str(checkdigit)

def check_account_mask_cd(mask, account, aggregate, demands_checkdigit):
	'''Checks whether an account code is correct against the mask format. 
	   "CD" means "check digit", that is, the account must come with check digit
	   (if the system is configured to demand that)
	   Returns an error message (empty string if no errors)'''
	
	if demands_checkdigit:
		if len(account) < 3:
			return "Account code too short to be valid"
		if account[-2] != checkdigit_separator:
			return "Checkdigit separator not found"
		# strips the checkdigit for the rest of account check
		cd = account[-1]
		account = account[:-2] 
		if checkdigit(account) != cd:
			return "Check digit mismatch - check the typed code"

	return check_account_mask_naked(mask, account, aggregate)

def check_account_mask_cdopt(mask, account, aggregate, demands_checkdigit):
	'''Checks whether an account code is correct. It can be supplied with or without a checkdigit,
	   but the checkdigit must be correct if it is there'''
	if account == strip_checkdigit(account):
		# no checkdigit
		return check_account_mask_naked(mask, account, aggregate)
	return check_account_mask_cd(mask, account, aggregate, demands_checkdigit)

def strip_checkdigit(account):
	'''strips checkdigit'''
	x = account.find(checkdigit_separator)
	if x < 0:
		return account
	return account[:x]

def check_account_mask_naked(mask, account, aggregate):
	'''Checks whether an account code is correct against the mask format. 
	   It is "naked" because the account must come without the check digit.
	   Returns an error message (empty string if no errors)'''

	if account.find(" ") >= 0:
		return "Unwanted spaces in account code"

	mask_err = check_chart_mask(mask)
	if mask_err:
		return "Account mask itself is bad (%s)" % mask_err

	mask_items = mask.split(mask_separator)
	account_items = account.split(mask_separator)

	if len(account_items) < 1:
		return "Account code is empty"

	if len(account_items) > len(mask_items):
		return "Account code does not match mask (too many levels)"

	if (not aggregate) and (len(mask_items) != len(account_items)):
		return "Non-aggregate accounts must fill all mask levels"
	
	for x in xrange(0, len(account_items)):
		mask_item = mask_items[x]
		account_item = account_items[x]

		if len(mask_item) > len(account_item):
			return "Level %d of account code too short" % (x+1)
		elif len(mask_item) < len(account_item):
			return "Level %d of account code too long" % (x+1)

		for y in xrange(0, len(mask_item)):
			if not mask_chars[mask_item[y]](account_item[y]):
				return "Level %d; wrong character at position %d" % (x+1, y+1)
		
	return ""

def check_chart_mask(mask):
	'''Checks whether an account *mask* (the mask itself) is ok. 
	   Returns an error message (empty if no error)'''

	if mask.find(" ") >= 0:
		return "Unwanted spaces in mask"

	items = mask.split(mask_separator)
	if len(items) < 2:
		return "There must be at least 2 levels in account mask"

	for item in items:
		if not len(item):
			return "Empty level in mask (2 consecutive dots)"
		for char in item:
			if char not in mask_chars:
				return "Character %s is not allowed in mask" % char

	return ""

def check_mask_9(c):
	return c.isdigit()

def check_mask_A(c):
	return c.isalpha() and c.isupper()

def check_mask_X(c):
	return check_mask_A(c) or check_mask_9(c)

mask_chars = {"9": check_mask_9, "A": check_mask_A, "X": check_mask_X}
mask_separator = "."
checkdigit_separator = "-"

