# -*- coding: UTF-8 -*-
"""Controller for the foreign currencies table."""

__all__ = ['currencies']

__author__    = "Christopher Arndt"
__revision__  = "$Rev$"
__date__      = "$Date$"
__license__   = "LGPL"

import time
import datetime
import urllib

from cStringIO import StringIO

import cherrypy
import turbogears

from turbogears import controllers, expose, identity, url, validate
from turbogears import validators as vldtrs

from sqlobject import sqlbuilder

from luca import model

from luca.model import Currency, conn, notfound, denied
from luca.simple_table import simple_table
from luca.widgets.currencies import *
from luca.commands.import_currencies import import_currencies

class currencies(simple_table):

    # SUBCLASS-SPECIFIC FUNCTIONS AND PARAMETERS ##########################

    validators = {"id": vldtrs.Int(if_invalid=0),
                  "code": vldtrs.UnicodeString(if_invalid=u""),
                  "name": vldtrs.UnicodeString(if_invalid=u""),
                  "autoDownload": vldtrs.Int(if_invalid=1),
                  "lite":  vldtrs.Int(if_invalid=0) \
                }

    def setup(self):
        ## special ACL that is accepted as valid permission
        #self.vips = ["ADMIN"]

        # data format for building JSON responses
        self.emptyrow_read = {"id": 0, "name": "", "code": "",
            "autoDownload": 1}
        self.emptyrow_write = self.emptyrow_read
        self.empty_intermediate = self.emptyrow_read

        self.formatters = dict(autoDownload=self.bool2string)

        # data for automated tests
        self.sample_data = [{},{},{},{}]
        self.invalid_data = [("code", ""), ("name", ""), ("autoDownload", "")]
        self.invalid_compound_new = [ dict(password="") ]
        self.unique_cols = ["code"]
        self.foreign_keys = []
        self.test_order = 9
        self.test_flags = ('cant_remove_id_1',)

        # error messages
        self.msg_unexpected = "ERROR: unexpected SQL error. Contact the system administrator"
        self.msg_notfound = "Currency not found, probably it has just been deleted"
        self.show_msg_ajax = False
        self.msg_added = "Currency %d has been added"
        self.msg_updated = "Currency has been updated"
        self.msg_removed = "Currency has been removed sucessfully"
        self.msg_validation = "Code and name must not be empty"
        self.msg_validation_crash = "Unexpected validation error: notify system administrator!"
        self.msg_integrity = "Integrity error: "
        self.msg_cannotremove = "Currency can not be removed"
        self.msg_cannotremove_admin = "The Administrator cannot be removed!"
        self.msg_unique = "Make sure that the currency code is unique"
        self.msg_legend_update = "Currencies: editing currency '%s'"
        self.msg_legend_add = "Currencies: adding new currency"

        self.table = Currency
        self.streamlined_add = True

        self.lite_options = [
            {"id": 0, "description": u"AJAX menu by default"},
            {"id": 1, "description": u"'Lite' (no Javascript) menu by default"}
        ]

    def row_name(self, d_row):
        name = ""
        if d_row["id"] > 0:
            name = u"%s (id %d)" % (d_row["code"], d_row["id"])
        else:
            name = d_row["code"]
        return name

    def intermediate_validate(self, d):
        err = ""
        if not d["code"]:
            err = "Currency code must not be empty"
        elif not d["name"]:
            err = "Currency name must not be empty"
        elif len(d["code"]) != 3: #pragma: no cover
            # paranoid - MySQL's varchar(240) will swallow up to 720 bytes
            err = "Currency code must be 3 characters long"
        if d["id"] < 0:
            if self.table.select(self.table.q.code == d["code"]).count() > 0:
                err = self.msg_integrity + self.msg_unique
        return err

    def data_legend(self, d):
        return d["code"]

    def auxiliary_form_data(self, form_data):
        form_data["lite_options"] = self.lite_options
        # FIXME: Hack to get around dumb tg_flash implementation
        form_data["tg_flash_class"] = "errmsg"

    # formatter methods
    def bool2string(self, val):
        if val:
            return "Yes"
        return "No"

    # EXPOSED METHODS #####################################################

    @expose(template="luca.templates.currencies")
    def index(self):
        data = self.base_index(orderBy=self.table.q.code)
        data["download_rates_form"] = download_rates_form
        data["download_rates_form_data"] = dict(
          action=url('/currency_rates/download_rates'),
          submit_text=_(u'Start download'),
          value = dict(startdate=datetime.date.today() - datetime.timedelta(7)))
        data["import_currencies_form"] = import_currencies_form
        data["import_currencies_form_data"] = dict(
          action=url('/currencies/import_currencies'),
          submit_text=_(u'Start import'))
        return data

    @expose(template="luca.templates.currencies_lite")
    def lite(self, **garbage):
        return self.base_index(orderBy=self.table.q.code)

    @expose("json")
    @validate(validators=validators)
    def remove(self, id, **garbage):
        return self.base_remove(id)

    @expose()
    @validate(validators=validators)
    def lite_remove(self, id, **garbage):
        return self.base_lite_remove(id)

    @expose("json")
    @validate(validators=validators)
    def fetch(self, id, **garbage):
        return self.base_fetch(id)

    @expose(template="luca.templates.currencies_lite_form")
    @validate(validators=validators)
    def lite_edit(self, id, **garbage):
        return self.base_lite_edit(id)

    @expose("json")
    @validate(validators=validators)
    def update(self, **form_data):
        return self.base_update(form_data)

    @expose(template="luca.templates.currencies_lite_form")
    @validate(validators=validators)
    def lite_commit(self,  **form_data):
        return self.base_lite_commit(form_data)

    @expose(template="luca.templates.message_lite", allow_json=True)
    @validate(form=import_currencies_form)
    def import_currencies(self, upload=None, url=None, tg_errors={}):
        data = {'err': []}
        file = None
        if 'upload' not in tg_errors:
            try:
                file = StringIO(upload.value)
            except:
                data['err'].append(_(u'Error reading uploaded file.'))
        elif url:
            try:
                file = urllib.urlopen(url)
            except:
                data['err'].append(_(u'Error retrieving URL.'))
        if file:
            try:
                ret = import_currencies(file, verbose=True)
            except Exception, exc:
                ret = 1
                if config('server.environment', 'development'):
                    data['err'].append(unicode(exc))
            else:
                data['msg'] = [_(u'Currency list sucessfully imported.'),
                  _(u'See server log for more information.')]
            if ret != 0:
                data['err'].append(_(u'Error importing currency list.'))
                data['err'].append(_(u'See server log for more information.'))
        else:
            data['err'].append(_(u'Please upload a file or specify a URL.'))
        data['title'] = _(u'Currencies import')
        data['back_to'] = ('/currencies/', _(u'Back to currencies'))
        return data
