# -*- coding: ISO-8859-1 -*-
"""The application's data model classes."""

import logging
import md5
import pickle
import time

from datetime import datetime, date, timedelta

import sqlobject
import cherrypy

from sqlobject import *

from turbogears import identity, config
from turbogears.database import PackageHub
from turbogears.identity import current

from luca import release

hub = PackageHub("luca")
__connection__ = hub
conn = hub
gconn = conn.getConnection()
backend = gconn.module
_testing_ = False
_testing_tag_ = 0

logger = logging.getLogger('luca.model')

class ModelError(Exception):
    # Model checking failure errors
    pass

class PermissionError(Exception):
    # Model checking failure errors
    pass

# an empty select is not a SQL "error", it is raised by SQLObject framework
notfound = (SQLObjectNotFound,)
integrity = (backend.IntegrityError, sqlobject.dberrors.IntegrityError)
modelerror = (ModelError,)
denied = (PermissionError,)
# catch-all
sqlerror = (SQLObjectNotFound, backend.Error, ModelError,
    backend.IntegrityError, sqlobject.dberrors.IntegrityError)

# These modules have individual/fine-grained control
modules = ["active_period", "balance_closing", "batches", "charts",
    "charts_accounts", "currencies", "currency_rates", "costcenters",
    "journals", "orgs", "stdjournals", "sys_classes", "sys_cmos"]

# Modules without individual control (ADMIN/REPORTS)
other_modules = [ "sys_users", "sys_groups", "sys_qlogs", "sys_usersXgroups",
    "sys_usersXorgs", "sys_groupsXperms", "report_journal", "report_balance",
    "report_income", "report_incomeavg", "report_cashflow" ]

function_points_nowild = ["read", "write.add", "write.edit", "write.remove"]

function_points       = ["*",   "read",       "write.*", "write.add",  "write.edit",         "write.remove"]
function_points_human = ["Any", "Read/Query", "Write",   "Write::Add", "Write::Update/Edit", "Write::Remove/Delete"]
function_points_sql   = ["%",   "read",       "write.%", "write.add",  "write.edit",         "write.remove"]

# These are meta-modules
# ALL_ORGS: can access all organizations, otherwise individual user<->org permissions are necessary
# REPORTS: all reports
# ADMIN: all system/admin tables (is NOT a super-user, does not give automatic access to e.g. reports!)
extra_modules = ["ALL_ORGS", "REPORTS", "ADMIN"]

# def to_unicode(s):
# 	if isinstance(s, str):
# 	    pass
# 	elif hasattr(s, '__unicode__'):
# 	    s = unicode(s)
# 	if isinstance(s, unicode):
# 	    s = s.encode('utf-8')
#       return s

soClasses = ('stdjournals', 'costcenters', 'classes', 'charts', 'orgs',
    'accounts', 'batches', 'journal', 'cmos', 'xuser', 'xgroup', 'xpermission',
    'log', 'group_permission', 'user_group', 'user_org', 'Visit',
    'VisitIdentity', 'Currency', 'CurrencyRate')

def touch():
    globs = globals()
    for tablename in soClasses:
        table = globs.get(tablename)
        #logger.debug('Creating table %s' % tablename)
        table.createTable(ifNotExists=True)

class Currency(SQLObject):
    """Foreign currency with name and code."""

    code = StringCol(notNone=True, alternateID=True, length=3)
    name = UnicodeCol(notNone=True, length=100)
    autoDownload = IntCol(notNone=True, default=1)

    fxrates = MultipleJoin('CurrencyRate')

    def _set_code(self, value):
        self._SO_set_code(value.upper())

    def download_rates(self, startdate=None, enddate=None, keep_manual=True):
        """Download rates for this currency for the period between
        startdate and enddate.

        startdate and enddate must be datatime.date types.
        """

        from util.fxrates_client import get_noon_rates, ParseError, NetworkError

        if enddate is None:
            enddate = date.today()
        if startdate is None:
            # default == 7 days before enddate
            startdate = enddate - timedelta(
              config.get('luca.fxrates.default_period', 7))

        msg = []
        admin = xuser.by_user_name('admin')
        try:
            rates = get_noon_rates(self.code, startdate, enddate)
        except (ParseError, NetworkError):
            # FIXME: should we handle the exception here?
            raise
        else:
            hub.begin()
            for rate, rdate in rates:
                try:
                    crate = CurrencyRate.selectBy(currency=self, date=rdate)[0]
                except IndexError:
                    CurrencyRate(currency=self, fxrate=rate, date=rdate,
                      manualOverride=0, user=admin)
                    msg.append('NEW: %s, rate=%s, date=%s' % (self.code,
                      rate, rdate))
                else:
                    if crate.fxrate != rate:
                        if not crate.manualOverride or not keep_manual:
                            crate.fxrate = rate
                            crate.manualOverride = False
                            user=admin
                            msg.append('UPDATE: %s, rate=%s, date=%s' % \
                              (self.code, rate, rdate))
                        else:
                            msg.append('SKIP: %s, rate=%s, date=%s' % \
                              (self.code, rate, rdate))
                    else:
                        msg.append('UNCHANGED: %s, rate=%s, date=%s' % \
                          (self.code, rate, rdate))
            hub.commit()
        return msg

class CurrencyRate(SQLObject):
    """Foreign currency exchange rate for a specific point in time."""

    date = DateCol(notNone=True, default=datetime.today)
    # x100 scaled
    fxrate = DecimalCol(notNone=True, size=16, precision=4)
    manualOverride = IntCol(notNone=True, default=1)

    currency = ForeignKey('Currency')
    user = ForeignKey('xuser')


class stdjournals(SQLObject):
    description = UnicodeCol(notNone=True)
    smalldesc = UnicodeCol(notNone=True)
    mobile = IntCol(notNone=True)
    inactive = IntCol(notNone=True, default=0, dbName="notactive")

    @classmethod
    def createTable(cls, *args, **kw):
        super(stdjournals, cls).createTable(*args, **kw)
        hub.commit()
        # it MUST have a id=1 empty row for journals with no standard journal
        try:
            stdjournals.get(1)
        except SQLObjectNotFound:
            stdjournals(description="", smalldesc="", mobile=0)

class costcenters(SQLObject):
    description = UnicodeCol(notNone=True)
    smalldesc = UnicodeCol(notNone=True)
    mobile = IntCol(notNone=True)
    inactive = IntCol(notNone=True, default=0, dbName="notactive")

    @classmethod
    def createTable(cls, *args, **kw):
        super(costcenters, cls).createTable(*args, **kw)
        hub.commit()
        # it MUST have a id=1 empty row for journals with no cost center
        try:
            costcenters.get(1)
        except SQLObjectNotFound:
            costcenters(description="", smalldesc="", mobile=0)

class classes(SQLObject):
    oid = IntCol(notNone=True, alternateID=True)
    description = UnicodeCol(notNone=True) # if it is an income account. 0=assets/liabs., 1=income/expense
    is_income = IntCol(notNone=True) # 0=cr 1=db
    normal_balance = IntCol(notNone=True)

    @classmethod
    def createTable(cls, *args, **kw):
        super(classes, cls).createTable(*args, **kw)
        hub.commit()
        # it MUST have one and only one row
        if classes.select().count() == 0:
            # empty table, let's populate it
            classes(oid=10, description="Assets", is_income=0, normal_balance=1)
            classes(oid=20, description="Liabilities", is_income=0, normal_balance=0)
            classes(oid=30, description="Equity", is_income=0, normal_balance=0)
            classes(oid=40, description="Revenue", is_income=1, normal_balance=0)
            classes(oid=50, description="Expenses", is_income=1, normal_balance=1)

class charts(SQLObject):
    description = UnicodeCol(notNone=True)
    # 6 x 20 chars due to UTF-8 expansion effect
    mask = UnicodeCol(length=120, notNone=True)
    # monotonic version to watch for changes
    version = IntCol(notNone=True, default=0)
    # whether the account tree is ok
    clean = IntCol(notNone=True, default=0)
    accounts = MultipleJoin('accounts', joinColumn='chart_id')

    def bump_version(self):
        self.set(version=(self.version+1))

class orgs(SQLObject):
    chart = ForeignKey('charts', notNone=True, notNull=True)
    name = UnicodeCol(notNone=True)
    active_d1 = DateCol(notNone=True, default='1970-01-01')
    active_d2 = DateCol(notNone=True, default='1970-01-01')
    remarks = UnicodeCol(notNone=True)
    # monotonic increasing to watch for changes
    version = IntCol(notNone=True, default=0)

    def bump_version(self):
        self.set(version=(self.version+1))

class accounts(SQLObject):
    chart = ForeignKey('charts', notNone=True, notNull=True)
    xclass = ForeignKey('classes', notNone=True, notNull=True)
    stdjournal = ForeignKey('stdjournals', notNone=True, notNull=True)
    code = UnicodeCol(length=120, notNone=True) # 6 x 20 unicode chars
    # I did not use a ForeignKey here because a chart tree may be broken at
    # times. Post-checking is needed before report calculation
    parent = UnicodeCol(length=120, notNone=True) # 6 x 20 unicode chars (utf-8 is culprit :)
    description = UnicodeCol(notNone=True)
    smalldesc = UnicodeCol(notNone=True)
    aggregate = IntCol(notNone=True)
    mobile = IntCol(notNone=True)
    inactive = IntCol(notNone=True, dbName="notactive")
    x = DatabaseIndex('chart', 'code', unique=True)

class batches(SQLObject):
    org = ForeignKey('orgs', notNone=True, notNull=True)
    date1 = DateCol(notNone=True)
    date2 = DateCol(notNone=True)
    description = UnicodeCol(notNone=True)
    # x100 scaled
    total = DecimalCol(size=16, precision=0, notNone=True)
    closed = IntCol(notNone=True)
    adjust = IntCol(notNone=True)
    # monotonic increasing to watch for changes
    version = IntCol(notNone=True, default=0)
    journals = MultipleJoin('journal', joinColumn='batch_id')

    def bump_version(self):
        self.set(version=(self.version+1))

class journal(SQLObject):
    org = ForeignKey('orgs', notNone=True, notNull=True)
    batch = ForeignKey('batches', notNone=True, notNull=True)
    account = ForeignKey('accounts', notNone=True, notNull=True)
    stdjournal = ForeignKey('stdjournals', notNone=True, notNull=True)
    costcenter = ForeignKey('costcenters', notNone=True, notNull=True)

    day = DateCol(notNone=True, dbName="tday")
    description = UnicodeCol(notNone=True)
    value = DecimalCol(size=16, precision=0, notNone=True, dbName="tvalue") # x100 scaled

    transa = IntCol(notNone=False) # Transaction ID: relates entry to other systems
    cancelled = IntCol(notNone=True, default=0)

    hash = UnicodeCol(length=240, notNone=True) # 40 x 6

    h = DatabaseIndex('hash', unique=True)
    a = DatabaseIndex('org', 'day')
    b = DatabaseIndex('transa')

def journal_hash(row):
    s = u"luca001 %d %d %d %d %d %s %s %.2f %f" % \
        (row["org"], row["batch"], row["account"], row["stdjournal"],
        row["costcenter"], row["description"], row["day"], row["value"],
        time.time())
    return md5.md5(s.encode("utf-8")).hexdigest()

class cmos(SQLObject):
	batch_zero = IntCol(notNone=True, default=1)
	batch_open = IntCol(notNone=True, default=1)
	accounts_checkdigit = IntCol(notNone=True, default=1)

	# all TABLE_version are monotonic counters to watch for changes in the TABLE
	stdjournals_version = IntCol(notNone=True, default=0)
	costcenters_version = IntCol(notNone=True, default=0)
	orgs_version = IntCol(notNone=True, default=0)
	charts_version = IntCol(notNone=True, default=0)
	xuser_version = IntCol(notNone=True, default=0)
	xgroup_version = IntCol(notNone=True, default=0)

	# this is the true database version
	database_version = UnicodeCol(notNone=True, default=release.version)

	@classmethod
	def createTable(cls, *args, **kw):
		super(cmos, cls).createTable(*args, **kw)
		hub.commit()
		# it MUST have one and only one row
		try:
			cmos.get(1)
		except SQLObjectNotFound:
			cmos() 
		if cmos.select().count() > 1:
			raise "DATABASE ASSERTION ERROR: More than one row in cmos table!"  #pragma: no cover

	@classmethod
	def bump_version(cls, table_name):
		o = cmos.get(1)
		p = "%s_version" % table_name
		q = getattr(o, p) + 1
		o.set(**{p: q})

	@classmethod
	def get_version(cls, table_name):
		o = cmos.get(1)
		p = "%s_version" % table_name
		q = getattr(o, p) + 0
		return q

# Generated by TurboGears

class xgroup(SQLObject):
    """
    An ultra-simple group definition.
    """

    group_name = UnicodeCol( length=96, alternateID=True, alternateMethodName="by_group_name" ) # 16 x 6 (unicode -> utf8)
    display_name = UnicodeCol()
    created = DateTimeCol( default=datetime.now )

    # collection of all users belonging to this group
    users = RelatedJoin( "xuser", intermediateTable="user_group",
                        joinColumn="group_id", otherColumn="user_id",
			createRelatedTable=False)

    # collection of all permissions for this group
    permissions = RelatedJoin( "xpermission", joinColumn="group_id",
                              intermediateTable="group_permission",
                              otherColumn="permission_id",
			      createRelatedTable=False)


class xuser(SQLObject):
    """
    Reasonably basic User definition. Probably would want additional attributes.
    """

    user_name = UnicodeCol( length=96, alternateID=True, alternateMethodName="by_user_name" )  # 16 x 6
    email_address = UnicodeCol( length=250, alternateID=True, alternateMethodName="by_email_address" ) 
    display_name = UnicodeCol( )
    password = UnicodeCol(length=240, dbName='pwd')
    created = DateTimeCol( default=datetime.now )
    lite = IntCol(notNone=True, default=0)
    locale = UnicodeCol(notNone=True, default=".,Y/m/d")

    # groups this user belongs to
    groups = RelatedJoin( "xgroup", intermediateTable="user_group",
                         joinColumn="user_id", otherColumn="group_id",
			 createRelatedTable=False)

    # organizations this user can manipulate
    orgs = RelatedJoin( "orgs", intermediateTable="user_org",
                         joinColumn="user_id", otherColumn="org_id",
			 createRelatedTable=False)

    def _get_permissions( self ):
        perms = set()
        for g in self.groups:
            perms = perms | set(g.permissions)
        return perms

    def _set_password( self, cleartext_password ):
        "Runs cleartext_password through the hash algorithm before saving."
        hash = identity.encrypt_password(cleartext_password)
        self._SO_set_password(hash)

    def set_password_raw( self, password ):
        "Saves the password as-is to the database."
        self._SO_set_password(password) #pragma: no cover

    @classmethod
    def createTable(cls, *args, **kw):
	super(xuser, cls).createTable(*args, **kw)
	hub.commit()
	try:
		adm = xuser.by_user_name("admin")
	except SQLObjectNotFound:
		adm = xuser(user_name="admin", email_address="root@localhost",
				display_name="Administrator", password="admin")

	xgroup.createTable(ifNotExists=True)
	user_group.createTable(ifNotExists=True)
	hub.commit()
	try:
		gadm = xgroup.by_group_name("admin")
	except SQLObjectNotFound:
		gadm = xgroup(group_name="admin", display_name="Administrators")

	xpermission.createTable(ifNotExists=True)
	group_permission.createTable(ifNotExists=True)
	hub.commit()
	try:
		perm = xpermission.by_permission_name("*")
	except SQLObjectNotFound:
		perm = xpermission(permission_name="*", description="g0d")

	hub.commit()
	if gadm not in adm.groups:
		adm.addXgroup(gadm)

	hub.commit()
	if perm not in gadm.permissions:
		perm.addXgroup(gadm)

	hub.commit()
	for module in modules:
		for function_point in function_points:
			name = module+"."+function_point
			try:
				xpermission.by_permission_name(name)
			except SQLObjectNotFound:
				xpermission(permission_name=name, description=name)

	for module in extra_modules:
		name = module
		try:
			xpermission.by_permission_name(name)
		except SQLObjectNotFound:
			xpermission(permission_name=name, description=name)


class log(SQLObject):
	module = UnicodeCol(notNone=True)
	function = UnicodeCol(notNone=True)
	description = UnicodeCol(notNone=True)
	err = UnicodeCol(notNone=True)
	user = ForeignKey('xuser', notNone=True, notNull=True)
	tmstamp = DateTimeCol(notNone=True, default = datetime.now)
	c = DatabaseIndex('tmstamp')


class xpermission(SQLObject):
    permission_name = UnicodeCol(length=50, alternateID=True, alternateMethodName="by_permission_name")
    description = UnicodeCol()
    
    hub.commit()

    groups = RelatedJoin( "xgroup",
                        intermediateTable="group_permission",
                         joinColumn="permission_id",
                         otherColumn="group_id",
			 createRelatedTable=False )

# These would be automatically created when xuser/xgroup/xpermission are created,
# but we need a unique index that SQLObject does not create by itself

class user_org(SQLObject):
	user = ForeignKey('xuser', notNull=True, notNone=True)
 	org = ForeignKey('orgs', notNull=True, notNone=True)
	x = DatabaseIndex('org', 'user', unique=True)

class group_permission(SQLObject):
	group = ForeignKey('xgroup', notNull=True, notNone=True)
 	permission = ForeignKey('xpermission', notNull=True, notNone=True)
	x = DatabaseIndex('group', 'permission', unique=True)

class user_group(SQLObject):
 	group = ForeignKey('xgroup', notNull=True, notNone=True)
	user = ForeignKey('xuser', notNull=True, notNone=True)
	x = DatabaseIndex('group', 'user', unique=True)

class Visit(SQLObject):
 	class sqlmeta:
 		table="visit"

 	visit_key= StringCol( length=40, alternateID=True, alternateMethodName="by_visit_key" )
 	created= DateTimeCol( default=datetime.now )
 	expiry= DateTimeCol()

	def lookup_visit( cls, visit_key ): #pragma: no cover
		try:
 			return cls.by_visit_key( visit_key )
 		except SQLObjectNotFound:
 			return None
 	lookup_visit= classmethod(lookup_visit)

class VisitIdentity(SQLObject):
	visit_key = StringCol(length=40, alternateID=True, alternateMethodName="by_visit_key")
	user_id = IntCol()


fake_session = {}

def session():
	# Returns a CherryPy persistent session object.
	# In the absense of that feature, returns a global dict.
	try:
		sess = cherrypy.session
		sess["_probe"] = 1
	except AttributeError: #pragma: no cover
		global fake_session
		sess = fake_session
	if 'lite' not in sess:
		if current.user:
			sess['lite'] = current.user.lite
		else: #pragma: no cover
			sess['lite'] = 1
	return sess

################ ACLs (security)

perm_messages = {"LOGIN": "Please log in to have access to this operation", \
		 "ACL": "You do not have permission to do this task", \
		 "ORG": "You do not have permission to access this organization",
		 "ERR": "SQL error while trying to check permissions"}

lastlog = None

class permission(object):
	MASK_DONE = 1
	MASK_FAIL = 2
	MASK_DENIED = 4

	def _write(self):
		if _testing_:
			self.description = u"%d %s" % (_testing_tag_, self.description)
		try:
			global lastlog
			lastlog = log(module=self.module, function=self.function, description=self.description,
						user=current.user, err=self.logerr)
		except sqlerror, e:  #pragma: no cover
			# Should not happen, deserves low-level logging
			if current.user is None:
				# not logged in
				luca.controllers.turbogears_log.error("Could not write Luca log: %s (%s)" % (str(e.args),
									self.module+" (Not logged in) "+self.function))
			else:
				luca.controllers.turbogears_log.error("Could not write Luca log: %s (%s)" % (str(e.args),
									self.module+" "+current.user.user_name+" "+self.function))

	def fail(self, logerr="Error not specified (probably SQL)"):
		if not self.burnt:
			self.burnt = 1
			self.logerr = logerr
			if not (self.logmask & permission.MASK_FAIL):
				self._write()

	def done(self, *extradata):
		if not self.burnt:
			self.burnt = 1
			if extradata:
				self.description += " "
				self.description += u" ".join([unicode(x) for x in extradata])
			if not (self.logmask & permission.MASK_DONE):
				self._write()

	def _denied(self, errcode):
		self.burnt = 1
		self.granted = ""
		self.err = perm_messages[errcode]
		if errcode != "LOGIN":
			self.logerr = "AUTH "+errcode
			if not (self.logmask & permission.MASK_DENIED):
				self._write()

	acl_cache = {}

	def acl_explode(self, function_point, vips):
		'''Calculates all possible fulfilling ACLs for the given function point, and return
			then as a SQL SELECT expression string'''

		cache_key = ".".join(vips) + "/" + function_point

		if cache_key in permission.acl_cache:
			acls = permission.acl_cache[cache_key]
		else:
			# Creates a list of wildcarded ACLs that match the required function point
			# e.g. sys_users.write.remove becomes ["*", "sys_users.*", "sys_users.write.*", ...]
			acls = {"*": 1}
			acl = ""
			for token in function_point.split(".")[:-1]:
				acl += token
				acls[acl + ".*"] = 1
				acl += "."
			acls[function_point] = 1

			for vip in vips:
				acls[vip] = 1

			permission.acl_cache[cache_key] = acls
		return acls


	def acl(self, function_point, vips, org):
		'''Calculates if user has enough permissions for the given function point'''

		if not current.user: #pragma: no cover
			# Not logged in
			return "LOGIN"

		if "EVERYBODY" in vips:
			# any user can user this module
		 	return ""

		acls = self.acl_explode(function_point, vips)

		for acl in acls:
			if acl in current.permissions:
				break
		else:
			# no possible ACLs match
			return "ACL"

		if (org > 0) and ("ALL_ORGS" not in current.permissions):
			for candidate in current.user.orgs:
				if candidate.id == org:
					break
			else:
				# even "admin" needs ALL_ORGS specific grant, or grants to specific organizations
				# Of course, admin always can add those grants to itself via sys_groupsXperms
				return "ORG"

		return ""


	def __init__(self, module, function, org, logmask, *extra):
		'''
		Checks permissions and returns a log object

		Parameters:     module: the controller object that is requesting permission
					(controllers will just pass 'self')

					controller may have a "vips" property that tells which
					"special" permissions are accepted (like ADMIN or REPORTS)

				function: which function point is requesting permission inside the
					  module e.g. "write.add" or "read" or "write.remove"

				org: ID of the organization being manipulated. Operations not
					bound to any particular org. should pass zero.

				logmask: bit 0 LSB: if enabled, supresses "permission denied"
					 bit 1: supresses permission granted/done() transactions
					 bit 2: supresses permission granted/fail() transactions
					(NOTE: "user not logged in" error is never logged)

					logmask is useful to avoid an excess of log entries that would
					clutter log queries. Also, some "errors" are handled gracefully
					by controllers themselves, so there is no need to log a scarry
					"permission denied" message for an operation that did not fail.

				extra: one or more additional free-type parameters that will help to
					clarify the log. Tipically, the ID or equivalent unique field
					of the SQL row being manipulated should be passed here.

		How the client should use it:

		grant = permission(self, "", "write.add", 0, changed_id)
		if grant.err:
			print "Error: ", grant.err
			return
		try:
			# do something
			grant.done()
		except error:
			grant.fail()

		'''

		if function not in function_points_nowild: #pragma: no cover
			raise StandardError("Unknown function point %s" % function)

		self.module = module.__class__.__name__
		self.burnt = 0
		self.function = function
		self.description = u" ".join([unicode(x) for x in extra])
		self.logerr = "" # error to be logged
		self.err = "" # human-readable error message that can be shown e.g. in the Web page
		self.granted = "yes"
		self.logmask = logmask
		self.org = org
		self.vips = []

		if hasattr(module, "vips"):
			# special ("VIP") permissions
			self.vips = module.vips

		function_point = "%s.%s" % (module.__class__.__name__, function)
		try:
			err = self.acl(function_point, self.vips, self.org)
			if err:
				# Permission denied
				self._denied(err)
				raise PermissionError(err)

		except sqlerror:  #pragma: no cover
			# Should never happen, deserves double logging
			luca.controllers.turbogears_log.error("Could not check ACL permission: %s" % str(e.args))
			self._denied("ERR")
			raise PermissionError("SQL Error")


