# -*- coding: UTF-8 -*-
"""Widgets for the currencies and currency_rates pages."""

__all__ = [
    'DecimalValidator',
    'DateValidator',
    'DownloadRatesSchema',
    'DownloadRatesFormFields',
    'DownloadRatesForm',
    'download_rates_form',
    'ImportCurrenciesSchema',
    'import_currencies_form',
]

__author__    = "Christopher Arndt"
__revision__  = "$Rev$"
__date__      = "$Date$"
__license__   = "LGPL"

import datetime
import decimal

from turbogears import validators, widgets

from base import *

class DecimalValidator(validators.FancyValidator):
    """Converting validator that checks for/converts to Decimal type."""

    messages = {'invalid': 'Invalid literal for Decimal.'}

    def _to_python(self, value, state):
        try:
            if value == "NaN": raise ValueError
            value = decimal.Decimal(str(value))
        except (TypeError, ValueError, decimal.InvalidOperation):
            raise validators.Invalid(self.message('invalid', {}), value, state)
        return value

class DateValidator(validators.FancyValidator):
    """Converting validator that checks for/converts to date type."""

    messages = {'invalid': 'Invalid date string (format YYYY-MM-DD)'}

    def _to_python(self, value, state):
        try:
            y, m, d = [int(x) for x in value.split('-', 2)]
            value = datetime.date(y, m, d)
        except (TypeError, ValueError):
            raise validators.Invalid(self.message('invalid', {}), value, state)
        return value

class ValidDateRangeValidator(validators.FancyValidator):
    """Verifies that startdate is equal or less than enddate.

    To be used as a chained validator for DownloadRatesSchema.
    """

    messages = {
      'invalid': "Startdate must equal or less that end date."
    }

    def validate_python(self, value, state):
        try:
            if not value["startdate"] <= value["enddate"]:
                raise validators.Invalid(self.message('invalid', {}), value,
                  state)
        except TypeError:
            pass

class DownloadRatesSchema(validators.Schema):
    code = validators.String(max=3)
    startdate = DateValidator()
    enddate = DateValidator()
    keep_manual = validators.StringBool()
    chained_validators = [ValidDateRangeValidator]

class DownloadRatesFormFields(widgets.WidgetsList):
    code = widgets.HiddenField()
    startdate = widgets.CalendarDatePicker(label=_(u'Start date'),
      name='startdate', format='%Y-%m-%d')
    enddate = widgets.CalendarDatePicker(label=_(u'End date'),
      name='enddate', format='%Y-%m-%d')
    kepp_manual = widgets.CheckBox(label=_(u'Keep manually edited rates?'),
      name='keep_manual', default=True)

class DownloadRatesForm(widgets.TableForm):
    javascript = [widgets.JSLink('luca', 'javascript/download_rates.js')]
    submit = widgets.SubmitButton(attrs=dict(id='download_rates_form_submit'))

class ImportCurrenciesSchema(validators.Schema):
    url = validators.URL(add_http=False)
    upload = validators.FieldStorageUploadConverter()

class ImportCurrenciesForm(widgets.TableForm):
    javascript = [widgets.JSLink('luca', 'javascript/download_rates.js')]
    submit = widgets.SubmitButton(attrs=dict(id='import_currencies_form_submit'))

download_rates_form = DownloadRatesForm(
  fields = DownloadRatesFormFields(),
  name = 'download_rates_form',
  form_attrs = dict(id='download_rates_form'),
  validator = DownloadRatesSchema())

import_currencies_form = ImportCurrenciesForm(
    name = 'import_currencies_form',
    form_attrs = dict(id='import_currencies_form'),
    fields = [
        widgets.FileField(name='upload', label=_(u'Upload')),
        widgets.TextField(name='url', label=_(u'URL'))],
    validator = ImportCurrenciesSchema()
)
