# Copyright (c) 2005 INdT - Instituto Nokia de Tecnologia, Elvis Pfutzenreuter.
# All rights reserved. This software is released under the LGPL licence.

import md5
from mylib import *

import random
import json

def key_serialize(k):
	# converts to byte array
	k = k.encode("utf-8")
	return k
	
def key_unserialize(uk):
	# converts from bytearray to unicode
	uk = uk.decode("utf-8")
	return uk

def attr_serialize(o):
	s = json.write(o) # gets unicode or utf-8 strings and returns utf-8 stream
	return s

def attr_unserialize(s):
	o = json.read(s) # swallows utf-8 and returns unicode strings
	return o

class client_database(object):
	def keygen(self, data):
		raise Exception("Method should have been overriden")

	def chk_data(self, data, check_key):
		# Throws DatamodelError
		raise Exception("Method should have been overriden")

	def truncate(self):
		self.__dbm_trunc()

	def __dbm_trunc(self):
		self._memo = {}
		self.__dbm_close()
		self.__dbm = self.dbmmodule.open(self.dbmfile, "n")
		self.__dbm_close()
		self.__dbm = None

	def __dbm_open(self):
		self.__dbm_close()
		self.__dbm = self.dbmmodule.open(self.dbmfile, self.openparams)

	def __dbm_close(self):
		if self.__dbm is not None:
			self.__dbm.close()
			del self.__dbm
			self.__dbm = None

	def __init__(self, dbmmodule, dbmfile, fast=False):
		self.__dbm = None
		self.dbmmodule = dbmmodule
		self.dbmfile = dbmfile
		if fast:
			# Python for Series 60 e32dbm provision
			self.openparams = "cf"
		else:
			self.openparams = "c"
		self._memo = {}

		try:
			self.__dbm_open()

			for mkey in self.__dbm.keys():
				mdata = self.__dbm[mkey]

				key = key_unserialize(mkey)
				data = attr_unserialize(mdata)

				self.chk_data(data, 1)
				self._memo[key] = data
	
			self.__dbm_close()

		except self.dbmmodule.error, e:
			self.__dbm_trunc()
			raise DatamodelError("DBM low-level corruption", *e.args)

		except DatamodelError, e:
			self.__dbm_trunc()
			raise DatamodelError("DBM h-corruption", *e.args)

		except (ValueError, json.ReadException, json.WriteException), e:
			self.__dbm_trunc()
			raise DatamodelError("DBM m-corruption", *e.args)
			
	def append(self, data):
		self.chk_data(data, 0)
		if not data.has_key("key"):
			self.keygen(data)
		key = data["key"]
		if self._memo.has_key(key_serialize(key)):
			raise DatamodelError("Key %s already exists in database" % key)
		self._append_or_update(key, data)
		return key
	
	def update(self, data):
		self.chk_data(data, 1)
		key = data["key"]
		if not self._memo.has_key(key):
			raise DatamodelError("No key %s exists in database" % key)
		self._append_or_update(key, data)
		return key

	def _append_or_update(self, key, data):
		self._memo[key] = data

		try:
			# 'key' here is Unicode, which dbm does not like
			self.__dbm_open()
			self.__dbm[key_serialize(key)] = attr_serialize(data)
			self.__dbm_close()

		except self.dbmmodule.error, e:
			self.__dbm_trunc()
			raise DatamodelError("DBM a-corruption", *e.args)

	def remove(self, key):
		if not self._memo.has_key(key):
			raise DatamodelError("No key %s exists in database" % key)
		self._remove_silently(key)

	def _remove_silently(self, key):
		del self._memo[key]

		try:
			self.__dbm_open()
			del self.__dbm[key_serialize(key)]
			self.__dbm_close()

		except self.dbmmodule.error, e:
			self.__dbm_trunc()
			raise DatamodelError("DBM rm-corruption", *e.args)

	def has_key(self, key):
		return self._memo.has_key(key)

	def get_record(self, key):
		if not self._memo.has_key(key):
			raise DatamodelError("No key %s exists in database" % key)
		return self._memo[key].copy()

	def balance_line(self, bank):
		try:
			self.__dbm_open()

			for key in bank:
				data = bank[key]
				self.chk_data(data, 1)
				self._memo[key] = data
				self.__dbm[key_serialize(key)] = attr_serialize(data)

			for key in self._memo.keys():
				if not bank.has_key(key):
					del self._memo[key]
					del self.__dbm[key_serialize(key)]

			self.__dbm_close()

		except self.dbmmodule.error, e:
			self.__dbm_trunc()
			raise DatamodelError("DBM b-corruption", *e.args)
		except DatamodelError, e:
			self.__dbm_trunc()
			raise DatamodelError("DBM hl-corruption", *e.args)

	def zap(self):
		self.__dbm_trunc()

	def undefined(self, *params):
		raise Exception("This method is not defined for the class")

	def count(self):
		return len(self._memo)

	def record_list(self):
		'''Dangerous because exposes internal dictionary that can be changed'''
		return self._memo.values()

	def key_list(self):
		return self._memo.keys()

	def get_dict(self):
		d = {}
		for key in self._memo.keys():
			# Expensive but avoids internal dictionary exposure
			d[key] = self._memo[key].copy()
		return d

	def random_key(self):
		return random.choice(self._memo.keys())
	

class journal(client_database):

	def __init__(self, dbmmodule, dbmfile, fast=False):
		client_database.__init__(self, dbmmodule, dbmfile, fast)
		self.__update_pos()

	def keygen(self, data):
		hash = md5.new("%f%d" % (now(), data["value"]))
		data["key"] = unicode(hash.hexdigest())

	def chk_data(self, data, check_key):
		try:
			if data["type"] != u"journalentry":
				raise DatamodelError("Data record is of wrong type")

			try:
				x = data["date"] + 1
				x = data["value"] + 1
				x = data["timestamp"] + 1
			except TypeError, e:
				raise DatamodelError("Data record has corrupted value", *e.args)
		except KeyError, e:
			raise DatamodelError("Data record lacks a key", *e.args)
	
	def balance(self):
		diff = 0
		for reg in self.record_list():
			diff += reg["value"]
		return diff

	def get_by_pos(self, pos):
		# TODO: optimize
		for reg in self.record_list():
			if reg["pos"] == pos:
				return reg.copy()
		else:
			raise DatamodelError("Entry #%d not found" % pos)
	
	def _sort_order(self, x, y):
		r = cmp(self._memo[x]["date"], self._memo[y]["date"])
		if r == 0:
			r = cmp(self._memo[x]["timestamp"], self._memo[y]["timestamp"])
		return r

	def __update_pos(self):
		# TODO: optimize
		k = self.key_list()
		k.sort(lambda x, y: self._sort_order(x, y))
		n = 0
		for key in k:
			self._memo[key]["pos"] = n
			n += 1

	def append(self, data):
		key = client_database.append(self, data)
		self.__update_pos()
		return self.get_record(key)["pos"]

	def update(self, data):
		key = client_database.update(self, data)
		self.__update_pos()
		return self.get_record(key)["pos"]

	def remove(self, data):
		client_database.remove(self, data)
		self.__update_pos()

	balance_line = client_database.undefined
	

class accounts(client_database):

	def __init__(self, dbmmodule, dbmfile, fast=False):
		client_database.__init__(self, dbmmodule, dbmfile, fast)

	def get_list(self):
		l = []
		for key in self.key_list():
			l.append((key, self.get_record(key)["description"]))
		l.sort(lambda x, y: cmp(x[1].lower(), y[1].lower()))
		return l

	def chk_data(self, data, check_key):
		try:
			if data["type"] != u"account":
				raise DatamodelError("Account data record is corrupted (type)")
		except KeyError, e:
			raise DatamodelError("Data record lacks a key", *e.args)

	def keygen(self, data):
		if not data.has_key("key"):
			DatamodelError("We cannot generate keys for accounts")

	append = client_database.undefined
	append_or_update = client_database.undefined
	update = client_database.undefined
	remove = client_database.undefined

class stdjournals(client_database):

	def __init__(self, dbmmodule, dbmfile, fast=False):
		client_database.__init__(self, dbmmodule, dbmfile, fast)

	def chk_data(self, data, check_key):
		try:
			if data["type"] != u"stdjournal":
				raise DatamodelError("Stdjournal data record is corrupted (type)")
		except KeyError, e:
			raise DatamodelError("Data record lacks a key", *e.args)

	def keygen(self, data):
		if not data.has_key("key"):
			raise Exception("We cannot generate keys for stdjournals")

	append = client_database.undefined
	append_or_update = client_database.undefined
	update = client_database.undefined
	remove = client_database.undefined

class cmos(client_database):
	def allocate(self):
		if not self.has_key(u"cmos"):
			client_database.append(self, {"type": u"cmos", "org": 1, "accounts_version": 0,
					"stdjournals_version": 0, "server_url": "http://192.168.141.4:22222", 
					"server_bt": "00:11:22:33:44:55"})

	def truncate(self):
		client_database.truncate(self)
		self.allocate()

	def __init__(self, dbmmodule, dbmfile, fast=False):
		client_database.__init__(self, dbmmodule, dbmfile, fast)
		self.allocate()

	def get(self, name):
		return client_database.get_record(self, u"cmos")[name]

	def set(self, name, value):
		r = client_database.get_record(self, u"cmos")
		r[name] = value
		client_database.update(self, r)

	def chk_data(self, data, check_key):
		try:
			if data["type"] != u"cmos":
				raise DatamodelError("CMOS data record is corrupted (type)" + unicode(data))
			try:
				x = data["org"] + 1
				x = data["accounts_version"] + 1
				x = data["stdjournals_version"] + 1
			except AttributeError:
				raise DatamodelError("CMOS Data record is corrupted (elements)")

		except KeyError, e:
			raise DatamodelError("Data record lacks a key", *e.args)
			
	
	def keygen(self, data):
		if not data.has_key("key"):
			data["key"] = u"cmos"

	def get_record(self, dummy=None):
		return client_database.get_record(self, u"cmos")

	append = client_database.undefined
	append_or_update = client_database.undefined
	update = client_database.undefined
	remove = client_database.undefined

