# Copyright (c) 2005 INdT - Instituto Nokia de Tecnologia, Elvis Pfutzenreuter.
# All rights reserved. This software is released under the LGPL licence.

# Wire protocol level. We try not to deal with Unicode issues here; all dict keys
# are normal strings, the Unicode strings only appear for the client and server
# databases.

import md5
from mylib import *
import select
import json
import urllib

net_exceptions = ( IOError, )
net_exceptions_nonfatal = ( )

terminator = "\t\t\t\r\n"

def register_net_exceptions(klasses, klasses_nonfatal=[]):
	global net_exceptions
	global net_exceptions_nonfatal
	exception_list = []
	exception_list_nonfatal = []
	for klass in net_exceptions:
		exception_list.append(klass)
	for klass in net_exceptions_nonfatal:
		exception_list_nonfatal.append(klass)
	for klass in klasses:
		exception_list.append(klass)
	for klass in klasses_nonfatal:
		exception_list_nonfatal.append(klass)
	net_exceptions = tuple(exception_list)
	net_exceptions_nonfatal = tuple(exception_list_nonfatal)

def wait_one(rdsock, wrsock, fake = 0, nowrite = 0, timeout = 30):
	rdsocks = []
	wrsocks = []

	if rdsock:
		rdsocks = [rdsock]
	if wrsock:
		wrsocks = [wrsock]

	rd, wr = wait_many(rdsocks, wrsocks, fake, nowrite)

	if rd:
		rd = rd[0]
	if wr:
		wr = wr[0]
	
	return (rd, wr)

def wait_many(rdsocks, wrsocks, fake = 0, nowrite = 0, timeout = 30):
	'''May receive 1 socket per parameter, or a list of sockets'''
	
	if not rdsocks and not wrsocks:
		raise CommError("protocol.wait() called without need")

	if fake:
		# mocksock
		rd = []
		wr = wrsocks[:]
		for socket in rdsocks:
			if len(socket.recvbuffer) > 0 or not socket.active:
				rd.append(socket)
	elif nowrite:
		# Python for Nokia Series 60 provision: select() does not
		# support waiting for write clearance, so we always report
		# socket as writable
		wr = wrsocks[:]
		if wr:
			rd = []
		else:
			rd, wr_dummy, ex_dummy = select.select(rdsocks, [], [], timeout)
	else:
		rd, wr, ex_dummy = select.select(rdsocks, wrsocks, [], timeout)


	if not rd and not wr:
		raise CommError_Idle("Connection idle for %d secs - cancelled" % timeout)

	return (rd, wr)

class automata(object):

	def finish(self):
		try:
			if self.sk:
				self.sk.shutdown(1)
		except net_exceptions_nonfatal:
			pass
		except net_exceptions:
			pass

		try:
			if self.sk:
				self.sk.close()
				del self.sk
		except net_exceptions_nonfatal:
			pass
		except net_exceptions:
			pass

		self.send_buffer = ""
		self.recv_buffer = ""
		self.active = 0
		self.data_receipt = None

	def die(self, *errlist):
		self.finish()
		raise CommError("Comm", *errlist)

	def tick(self):
		return False

	def tick_read(self):
		nonfatal_error = False
		try:
			b = self.sk.recv(4096)
		except net_exceptions_nonfatal, e:
			b = ""
			nonfatal_error = True
		except net_exceptions, e:
			b = ""
			self.die("Socket", *e.args)

		if not b and not nonfatal_error:
			self.finish()
			return

		self.recv_buffer += b

		data = self.get_data()

		if data:
			if not self.data_receipt:
				raise CommError("No receipt for received data!")
			self.data_receipt(data)
		
	def tick_write(self):
		if not self.send_buffer:
			raise CommError("Internal error - tick_write - nothing to write")
		try:
			tosend = self.send_buffer
			if len(tosend) > 300:
				# provision to cheapo dongles: if you send too much data
				# the cheapo dongle at receiving side loses data :(
				tosend = tosend[0:300]
			written = self.sk.send(tosend)
		except net_exceptions_nonfatal:
			written = 0
			pass
		except net_exceptions, e:
			self.die("Socket", *e.args)

		self.send_buffer = self.send_buffer[written:]

	def sock_needs(self):
		rd = None
		wr = None
		if self.data_receipt:
			rd = self.sk
		if self.send_buffer:
			wr = self.sk
		return (rd, wr)

	def put_data(self, data):
		self.send_buffer += json.write(data) 
		self.send_buffer += terminator

	def chk_message(self, data):
		'''Should be extended by daughter classes'''
		return isinstance(data, dict)

	def get_data(self):
		data = None

		while self.recv_buffer.find(terminator) >= 0 and data is None:
			i = self.recv_buffer.find(terminator)
			line = self.recv_buffer[0:i]
			self.recv_buffer = self.recv_buffer[i+len(terminator):]

			if len(line) < 3:
				self.die("line too short")
				break

			dumped = line

			try:
				data = json.read(dumped)

			except (json.ReadException, json.WriteException, EOFError, ValueError), e:
				self.die("JSON decoding problem")
				data = None
				break

			if not self.chk_message(data):
				self.die("malformed message")
				data = None
				break

			if data.has_key("error"):
				self.die("Server error: %s" % data['error'])
				data = None
				break
			
		return data

	def __init__(self, sk):
		self.sk = sk
		self.recv_buffer = ""
		self.send_buffer = ""
		self.active = 1
		self.push = 0
		self.data_receipt = None
		self.name = "automata"

class client(automata):
	def chk_message(self, data):
		return automata.chk_message(self, data) and (data.has_key("response") or data.has_key("error"))

	def helo_push(self):
		d = { "request": "helo",  "org": self.cmos.get("org")}
		self.put_data( d )
		self.data_receipt = self.helo

	def helo(self, data):
		if data["response"] != "helo":
			self.die("helo: unexpected response")
			return

		self.helo_response = data

		# Ok we surrender control
		self.data_receipt = None

	def getaccounts_push(self):
		if self.helo_response["accounts_version"] == self.cmos.get("accounts_version"):
			# Nothing to see here, move along
			self.data_receipt = None
			return
	
		self.put_data( { "request": "getaccounts", } )
		self.data_receipt = self.getaccounts

	def getaccounts(self, data):
		if data["response"] != "getaccounts":
			self.die("getaccounts: unexpected response")
			return

		# balance_line may throw exception
		self.accounts.balance_line(data["payload"])
		self.cmos.set("accounts_version", self.helo_response["accounts_version"])

		# Ok we surrender control
		self.data_receipt = None

	def getstdjournals_push(self):
		if self.helo_response["stdjournals_version"] == self.cmos.get("stdjournals_version"):
			# Nothing to see here, move along
			self.data_receipt = None
			return

		self.put_data( { "request": "getstdjournals" } )
		self.data_receipt = self.getstdjournals

	def getstdjournals(self, data):
		if data["response"] != "getstdjournals":
			self.die("getstdjournals: unexpected response")
			return

		# balance_line may throw excepttion
		self.stdjournals.balance_line(data["payload"])
		self.cmos.set("stdjournals_version", self.helo_response["stdjournals_version"])

		# Ok we surrender control
		self.data_receipt = None

	def sendjournal_push(self):
		if self.journal.count() <= 0:
			self.data_receipt = None
			return

		self.put_data( { "request": "sendjournal", "payload": self.journal.get_dict() } )
		self.data_receipt = self.sendjournal

	def sendjournal(self, data):
		if data["response"] != "sendjournal":
			self.die("sendjournal: unexpected response")
			return

		# we assume that, if there was a response at all, journal sending was ok.
		self.data_receipt = None

	def confirmjournal_push(self):
		if self.journal.count() <= 0:
			self.data_receipt = None
			return

		self.put_data( { "request": "confirmjournal", "key": self.journal.random_key()} )
		self.data_receipt = self.confirmjournal

	def confirmjournal(self, data):
		if data["response"] != "confirmjournal":
			self.die("confirmjournal: unexpected response")
			return

		if not data["exists"]:
			self.die("Journal entries could not be sent safely")
			return

		# Ok we surrender control
		self.data_receipt = None

	def finished_push(self):
		self.journal.zap()
		# self.put_data( { "request": "finish" } )
		self.data_receipt = None

	def tick(self):
		if not self.active:
			return False
		elif self.data_receipt:
			return False
		elif self.send_buffer:
			return False
		elif self.tick_list:
			calltick = self.tick_list[0]
			del self.tick_list[0]
			calltick()
			return True
		else:
			self.finish()
			return True

	def __init__(self, sk, cmos, stdjournals, accounts, journal):
		automata.__init__(self, sk)
		self.name = "client"
		self.cmos = cmos
		self.journal = journal
		self.accounts = accounts
		self.stdjournals = stdjournals

		self.tick_list = [self.helo_push, self.getaccounts_push,
				  self.getstdjournals_push, self.sendjournal_push,
				  self.confirmjournal_push, self.confirmjournal_push,
				  self.finished_push]

def clientweb_default_url(cmos):
	return cmos.get("server_url")

def clientweb_set_default_url(cmos, url):
	return cmos.set("server_url", url)

class clientweb(client):
	def __init__(self, url, cmos, stdjournals, accounts, journal):
		client.__init__(self, None, cmos, stdjournals, accounts, journal)
		self.url = url

	def tick_read(self):
		data = self.get_data()
		self.recv_buffer = ""
		if data:
			if not self.data_receipt:
				raise CommError("No receipt for received data!")
			self.data_receipt(data)

	def tick_write(self):
		if not self.send_buffer:
			raise CommError("tick_write - internal error - nothing to write")
		params = urllib.urlencode({'blob': self.send_buffer})
		try:
			f = urllib.urlopen("%s/remote_mobile?%s" % (self.url, params))
			self.send_buffer = ""
			self.recv_buffer = f.read()
			self.recv_buffer += terminator # fools get_data() 
		except IOError, e:
			raise CommError("Web communication problem: %s" % str(e.args[1][1]))

